#
# Copyright (c) 2003 Art Haas
#
# This file is part of PythonCAD.
#
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# Transfer objects from one layer to another
#

import Generic.point
import Generic.circle
import Generic.arc
import Generic.conobject
import Generic.hcline
import Generic.vcline
import Generic.acline
import Generic.segjoint
import Generic.cline
import Generic.ccircle
import Generic.leader
import Generic.polyline
import Generic.text
import Generic.dimension
import Generic.layer
import Generic.image

def _remove_objects(objdict, from_layer):
    _objects = objdict.values()
    for _obj in _objects:
        _oid = id(_obj)
        if objdict[_oid] is not False:
            if isinstance(_obj, Generic.point.Point):
                from_layer.delObject(_obj)
                objdict[_oid] = False
            elif isinstance(_obj, (Generic.segment.Segment,
                                   Generic.cline.CLine)):
                if isinstance(_obj, Generic.segment.Segment):
                    _p1, _p2 = _obj.getEndpoints()
                else:
                    _p1, _p2 = _obj.getKeypoints()
                _pid = id(_p1)
                objdict[_pid] = False
                _pid = id(_p2)
                objdict[_pid] = False
                from_layer.delObject(_obj)
            elif isinstance(_obj, (Generic.circle.Circle,
                                   Generic.arc.Arc,
                                   Generic.ccircle.CCircle)):
                _cp = _obj.getCenter()
                _pid = id(_cp)
                objdict[_pid] = False
                if isinstance(_obj, Generic.arc.Arc):
                    for _ep in _obj.getEndpoints():
                        _lp = from_layer.find('point', _ep[0], _ep[1])
                        if _lp is None:
                            raise ValueError, "No Point at arc endpoint: " + str(_ep)
                        _pid = id(_lp)
                        objdict[_pid] = False
                from_layer.delObject(_obj)
            elif isinstance(_obj, (Generic.hcline.HCLine,
                                   Generic.vcline.VCLine,
                                   Generic.acline.ACLine)):
                _lp = _obj.getLocation()
                _pid = id(_lp)
                objdict[_pid] = False
                from_layer.delObject(_obj)
            elif isinstance(_obj, Generic.leader.Leader):
                _p1, _p2, _p3 = _obj.getPoints()
                _pid = id(_p1)
                objdict[_pid] = False
                _pid = id(_p2)
                objdict[_pid] = False
                _pid = id(_p3)
                objdict[_pid] = False
                from_layer.delObject(_obj)
            elif isinstance(_obj, Generic.polyline.Polyline):
                for _pt in _obj.getPoints():
                    _pid = id(_pt)
                    objdict[_pid] = False
                from_layer.delObject(_obj)
            elif isinstance(_obj, (Generic.segjoint.Chamfer,
                                   Generic.segjoint.Fillet)):
                for _seg in _obj.getSegments():
                    _sid = id(_seg)
                    for _pt in _seg.getEndpoints():
                        _pid = id(_pt)
                        objdict[_pid] = False
                    if objdict[_sid] is not False:
                        from_layer.delObject(_seg)
                        objdict[_sid] = False
            else:
                print "Unexpected deletion type %s" % type(_obj)
            objdict[_oid] = False
    for _oid, _obj in objdict.items():
        if _obj is not False:
            print "Skipped object: " + `_obj`
            print "id: " + _oid
            print "object: " + str(_obj)

def _adjust_dimensions(dimobjs, newobjs, to_layer):
    # print "in adjust_dimensions() ..."
    for _oid, _obj in dimobjs.items():
        # print "testing obj: " + `_obj`
        if _oid not in newobjs:
            raise KeyError, "Object not transfered: " + `_obj`
        _newobj = newobjs[_oid]
        _udict = {}
        for _uref in _obj.getUsers():
            _user = _uref()
            if _user is not None:
                _uid = id(_user)
                if _uid not in newobjs:
                    _udict[_uid] = _user
        _reset = True
        for _user in _udict.values():
            if not isinstance(_user, Generic.dimension.Dimension):
                _reset = False
                break
        # print "reset: " + str(_reset)
        if _reset:
            if isinstance(_obj, Generic.point.Point):
                for _dim in _udict.values():
                    if isinstance(_dim, Generic.dimension.LinearDimension):
                        _p1, _p2 = _dim.getDimPoints()
                        if _obj is _p1:
                            _dim.setP1(to_layer, _newobj)
                        elif _obj is _p2:
                            _dim.setP2(to_layer, _newobj)
                        else:
                            raise ValueError, "Point not used in linear dimension: " + `_dim`
                    elif isinstance(_dim, Generic.dimension.AngularDimension):
                        _vp, _p1, _p2 = _dim.getDimPoints()
                        if _obj is _vp:
                            _dim.setVertexPoint(to_layer, _newobj)
                        elif _obj is _p1:
                            _dim.setP1(to_layer, _newobj)
                        elif _obj is _p2:
                            _dim.setP2(to_layer, _newobj)
                        else:
                            raise ValueError, "Point not used in angular dimension: " + `_dim`
                    else:
                        raise TypeError, "Unexpected dim type for point: %s" % type(_dim)
            elif isinstance(_obj, Generic.circle.Circle): # and Arc
                for _dim in _udict.values():
                    if isinstance(_dim, Generic.dimension.RadialDimension):
                        _dim.setDimCircle(to_layer, _newobj)
                    else:
                        raise TypeError, "Unexpected dim type for circle/arc: %s" % type(_dim)
            else:
                raise TypeError, "Unexpected type: %s" % type(_obj)

def _used_in_dimension(obj):
    if isinstance(obj, (Generic.point.Point, Generic.circle.Circle)):
        for _uref in obj.getUsers():
            _user = _uref()
            if _user is not None:
                if isinstance(_user, Generic.dimension.Dimension):
                    return True
    return False

def transfer_objects(objlist, from_layer, to_layer):
    """Transfer objects from one layer to another.

transfer_objects(objlist, from_layer, to_layer)

objlist: A list of objects to transfer.
from_layer: A layer object containing the list of objects
to_layer: A layer object which will recieve the objects in the list
    """
    if not isinstance(objlist, list):
        raise TypeError, "Invalid object list: " + str(objlist)
    if not isinstance(from_layer, Generic.layer.Layer):
        raise TypeError, "Invalid layer: " + str(from_layer)
    if not isinstance(to_layer, Generic.layer.Layer):
        raise TypeError, "Invalid layer: " + str(to_layer)
    for _obj in objlist:
        if _obj not in from_layer:
            raise ValueError, "Object not in layer: " + `_obj`
    #
    # get the various subparts that will have to be transferred
    #
    _objdict = {}
    _dimobjs = {}
    _dtobjs = [] # dimension and text objects are handled specially
    while len(objlist):
        _obj = objlist.pop(0)
        # print "examining obj: " + `_obj`
        _oid = id(_obj)
        if _oid not in _objdict:
            _objdict[_oid] = _obj
        if isinstance(_obj, Generic.point.Point):
            if _used_in_dimension(_obj):
                _dimobjs[_oid] = _obj
        elif isinstance(_obj, (Generic.segment.Segment,
                               Generic.cline.CLine,
                               Generic.leader.Leader,
                               Generic.polyline.Polyline)):
            if isinstance(_obj, Generic.segment.Segment):
                _ptmeth = _obj.getEndpoints
            elif isinstance(_obj, Generic.cline.CLine):
                _ptmeth = _obj.getKeypoints
            elif isinstance(_obj, (Generic.leader.Leader,
                                   Generic.polyline.Polyline)):
                _ptmeth = _obj.getPoints
            for _pt in _ptmeth():
                _pid = id(_pt)
                if _pid not in _objdict:
                    _objdict[_pid] = _pt
                if _used_in_dimension(_pt):
                    _dimobjs[_pid] = _pt
        elif isinstance(_obj, (Generic.circle.Circle,
                               Generic.arc.Arc,
                               Generic.ccircle.CCircle)):
            _cp = _obj.getCenter()
            _cid = id(_cp)
            if _cid not in _objdict:
                _objdict[_cid] = _cp
            if _used_in_dimension(_cp):
                _dimobjs[_cid] = _cp
            if _used_in_dimension(_obj):
                _dimobjs[_oid] = _obj
            if isinstance(_obj, Generic.arc.Arc):
                for _ep in _obj.getEndpoints():
                    _lp = from_layer.find('point', _ep[0], _ep[1])
                    if _lp is None:
                        raise ValueError, "No Point at arc endpoint: " + str(_ep)
                    _lid = id(_lp)
                    if _lid not in _objdict:
                        _objdict[_lid] = _lp
                    if _used_in_dimension(_lp):
                        _dimobjs[_lid] = _lp
        elif isinstance(_obj, (Generic.hcline.HCLine,
                               Generic.vcline.VCLine,
                               Generic.acline.ACLine)):
            _lp = _obj.getLocation()
            _lid = id(_lp)
            if _lid not in _objdict:
                _objdict[_lid] = _lp
            if _used_in_dimension(_lp):
                _dimobjs[_lid] = _lp
        elif isinstance(_obj, (Generic.segjoint.Chamfer,
                               Generic.segjoint.Fillet)):
            objlist.extend(_obj.getSegments())
        elif isinstance(_obj, (Generic.dimension.LinearDimension,
                               Generic.dimension.RadialDimension,
                               Generic.dimension.AngularDimension,
                               Generic.text.TextBlock)):
            del _objdict[_oid]
            _dtobjs.append(_obj)
        else:
            print "Skipping object type %s" % type(_obj)
    #
    # now start moving the things between layers
    #
    _newobjs = {}
    _sjoints = []
    for _oid, _obj in _objdict.items():
        if _oid in _newobjs:
            continue
        if isinstance(_obj, Generic.point.Point):
            _x, _y = _obj.getCoords()
            _pt = to_layer.find('point', _x, _y)
            if _pt is None:
                _pt = _obj.clone()
                to_layer.addObject(_pt)
            _newobjs[_oid] = _pt
        elif isinstance(_obj, (Generic.segment.Segment,
                              Generic.cline.CLine)):
            _seg = False
            if isinstance(_obj, Generic.segment.Segment):
                _p1, _p2 = _obj.getEndpoints()
                _seg = True
            else:
                _p1, _p2 = _obj.getKeypoints()
            _pid = id(_p1)
            if _pid in _newobjs:
                _np1 = _newobjs[_pid]
            else:
                _x, _y = _p1.getCoords()
                _np1 = to_layer.find('point', _x, _y)
                if _np1 is None:
                    _np1 = _p1.clone()
                    to_layer.addObject(_np1)
                _newobjs[_pid] = _np1
            _pid = id(_p2)
            if _pid in _newobjs:
                _np2= _newobjs[_pid]
            else:
                _x, _y = _p2.getCoords()
                _np2 = to_layer.find('point', _x, _y)
                if _np2 is None:
                    _np2 = _p2.clone()
                    to_layer.addObject(_np2)
                _newobjs[_pid] = _np2
            if _seg:
                _newobj = _obj.clone()
                _newobj.setP1(_np1)
                _newobj.setP2(_np2)
                _newobj.reset()
            else:
                _newobj = Generic.cline.CLine(_np1, _np2)
            to_layer.addObject(_newobj)
            _newobjs[_oid] = _newobj
        elif isinstance(_obj, (Generic.leader.Leader,
                               Generic.polyline.Polyline)):
            _leader = False
            _npts = []
            if isinstance(_obj, Generic.leader.Leader):
                _leader = True
            for _pt in _obj.getPoints():
                _pid = id(_pt)
                if _pid in _newobjs:
                    _npt= _newobjs[_pid]
                else:
                    _x, _y = _pt.getCoords()
                    _npt = to_layer.find('point', _x, _y)
                    if _npt is None:
                        _npt = _pt.clone()
                        to_layer.addObject(_npt)
                    _newobjs[_pid] = _npt
                _npts.append(_npt)
            _newobj = _obj.clone()
            if _leader:
                _newobj.setP1(_npts[0])
                _newobj.setP2(_npts[1])
                _newobj.setP3(_npts[2])
            else:
                for _i in range(len(_npts)):
                    _newobj.setPoint(_i, _npts[_i])
            _newobj.reset()
            to_layer.addObject(_newobj)
            _newobjs[_oid] = _newobj
        elif isinstance(_obj, (Generic.hcline.HCLine,
                               Generic.vcline.VCLine,
                               Generic.acline.ACLine)):
            _pt = _obj.getLocation()
            _pid = id(_pt)
            if _pid in _newobjs:
                _npt = _newobjs[_pid]
            else:
                _x, _y = _pt.getCoords()
                _npt = to_layer.find('point', _x, _y)
                if _npt is None:
                    _npt = _pt.clone()
                    to_layer.addObject(_npt)
                _newobjs[_pid] = _npt
            if isinstance(_obj, Generic.hcline.HCLine):
                _newobj = Generic.hcline.HCLine(_npt)
            elif isinstance(_obj, Generic.vcline.VCLine):
                _newobj = Generic.vcline.VCLine(_npt)
            elif isinstance(_obj, Generic.acline.ACLine):
                _newobj = Generic.acline.ACLine(_npt, _obj.getAngle())
            else:
                raise TypeError, "Unexpected dimension type: " + type(_obj)
            to_layer.addObject(_newobj)
            _newobjs[_oid] = _newobj
        elif isinstance(_obj, (Generic.circle.Circle,
                               Generic.arc.Arc,
                               Generic.ccircle.CCircle)):
            _pt = _obj.getCenter()
            _pid = id(_pt)
            if _pid in _newobjs:
                _npt = _newobjs[_pid]
            else:
                _x, _y = _pt.getCoords()
                _npt = to_layer.find('point', _x, _y)
                if _npt is None:
                    _npt = _pt.clone()
                    to_layer.addObject(_npt)
                _newobjs[_pid] = _npt
            if isinstance(_obj, Generic.ccircle.CCircle):
                _newobj = Generic.ccircle.CCircle(_npt, _obj.getRadius())
            else:
                #
                # arc endpoints will get added because they are
                # in the _objdict dictionary
                #
                _newobj = _obj.clone()
                _newobj.setCenter(_npt)
                _newobj.reset()
            to_layer.addObject(_newobj)
            _newobjs[_oid] = _newobj
        elif isinstance(_obj, (Generic.segjoint.Chamfer,
                               Generic.segjoint.Fillet)):
            _s1, _s2 = _obj.getSegments()
            _is1 = id(_s1)
            _is2 = id(_s2)
            if _is1 in _newobjs and _is2 in _newobjs:
                _ns1 = _newobjs[_is1]
                _ns2 = _newobjs[_is2]
                if isinstance(_obj, Generic.segjoint.Chamfer):
                    _newobj = Generic.segjoint.Chamfer(_ns1, _ns2,
                                                       _obj.getLength())
                else:
                    _newobj = Generic.segjoint.Fillet(_ns1, _ns2,
                                                      _obj.getRadius())
                _newobj.setStyle(_obj.getStyle())
                _newobj.setLinetype(_obj.getLinetype())
                _newobj.setColor(_obj.getColor())
                _newobj.setThickness(_obj.getThickness())
                to_layer.addObject(_newobj)
            else:
                _sjoints.append(_obj)
        else:
            print "Skipping object type %s in objdict" % type(_obj)
    #
    # handle any chamfers and fillets now that all the segments
    # they join should have cloned copies in the 'to' layer
    #
    if len(_sjoints):
        for _obj in _sjoints:
            _s1, _s2 = _obj.getSegments()
            _is1 = id(_s1)
            if _is1 not in _newobjs:
                raise KeyError, "Segjoint s1 segment not transfered"
            _is2 = id(_s2)
            if _is2 not in _newobjs:
                raise KeyError, "Segjoint s2 segment not transfered"
            _ns1 = _newobjs[_is1]
            _ns2 = _newobjs[_is2]
            if isinstance(_obj, Generic.segjoint.Chamfer):
                _newobj = Generic.segjoint.Chamfer(_ns1, _ns2,
                                                   _obj.getLength())
            else:
                _newobj = Generic.segjoint.Fillet(_ns1, _ns2,
                                                  _obj.getRadius())
            _newobj.setStyle(_obj.getStyle())
            _newobj.setLinetype(_obj.getLinetype())
            _newobj.setColor(_obj.getColor())
            _newobj.setThickness(_obj.getThickness())
            to_layer.addObject(_newobj)
    #
    # adjust dimensions attached to objects that may have been
    # transfered between the layers
    #
    if len(_dimobjs):
        _adjust_dimensions(_dimobjs, _newobjs, to_layer)
    #
    # transfer any dimensions and text blocks
    #
    # text blocks do not reference any points so they can be transfered
    # at will, but as dimensions references points, circle, or arcs, moving
    # them has to be done by first deleting them from the layer they
    # currently sit in (and thusly diconnecting them from the points or
    # circlular objects they use) and then adding them to the new layer
    # if they are first added to the new layer and then deleted from the
    # old, the points and circles dimensions refer to can lose track of the
    # fact that there is a dimension bound to themselves
    #
    if len(_dtobjs):
        for _obj in _dtobjs:
            from_layer.delObject(_obj)
            to_layer.addObject(_obj)
    #
    # now remove the objects from the layer they where they
    # are currently stored
    #
    _remove_objects(_objdict, from_layer)



#
# Copyright (c) 2002, 2003, 2004 Art Haas
#
# This file is part of PythonCAD.
#
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# classes for line segments
#
# TODO: Look at using weak references

from __future__ import generators

import math

import graphicobject
import entity
import tolerance
import style
import linetype
import color
import point
import util
import quadtree

class Segment(graphicobject.GraphicObject):
    """A class representing a line segment.

A Segment has two attributes.

p1: A Point object representing the first end point
p2: A Point object representing the second end point

A Segment has the following methods:

getEndpoints(): Return the two endpoints of the Segment.
{get/set}P1: Get/Set the Segment first endpoint.
{get/set}P2: Get/Set the Segment second endpoint.
move(): Move the Segment.
length(): Get the Segment length.
getCoefficients(): Return the segment as ax + by + c = 0
getProjection(): Project a coordinate on to the Segment
mapCoords(): Test if a coordinate pair is within some distance to a Segment.
inRegion(): Test if the Segment is visible in some area.
clone(): Make an identical copy of a Segment.
    """
    messages = {
        'moved' : True,
        'endpoint_changed' : True
        }

    __defstyle = style.Style(u'Default Segment Style',
                            linetype.Linetype(u'Solid', None),
                            color.Color(0xffffff),
                            1.0)

    def __init__(self, p1, p2, st=None, lt=None, col=None, th=None, **kw):
        """Initialize a Segment object.

Segment(p1, p2[, st, lt, col, th])

p1: Segment first endpoint - may be a Point or a two-item tuple of floats.
p2: Segment second endpoint - may be a Point or a two-item tuple of floats.

The following arguments are optional:

st: A Style object
lt: A Linetype object that overrides the linetype in the Style.
col: A Color object that overrides the color in the Style.
th: A float that overrides the line thickness in the Style.
        """

        _p1 = p1
        if not isinstance(_p1, point.Point):
            _p1 = point.Point(p1)
        _p2 = p2
        if not isinstance(_p2, point.Point):
            _p2 = point.Point(p2)
        if _p1 is _p2:
            raise ValueError, "Segments cannot have identical endpoints."
        _st = st
        if _st is None:
            _st = Segment.__defstyle
        graphicobject.GraphicObject.__init__(self, _st, lt, col, th, **kw)
        self.__p1 = _p1
        _p1.connect('moved', self, Segment.movePoint)
        _p1.storeUser(self)
        self.__p2 = _p2
        _p2.connect('moved', self, Segment.movePoint)
        _p2.storeUser(self)

    def __str__(self):
        return "Segment: %s to %s" % (self.__p1, self.__p2)

    def __eq__(self, obj):
        """Compare a Segment to another for equality.
        """
        if not isinstance(obj, Segment):
            return False
        if obj is self:
            return True
        _sp1 = self.__p1
        _sp2 = self.__p2
        _op1, _op2 = obj.getEndpoints()
        if ((_sp1 == _op1) and (_sp2 == _op2)):
            _val = True
        elif ((_sp1 == _op2) and (_sp2 == _op1)):
            _val = True
        else:
            _val = False
        return _val

    def __ne__(self, obj):
        """Compare a Segment to another for inequality.
        """
        if not isinstance(obj, Segment):
            return True
        if obj is self:
            return False
        _sp1 = self.__p1
        _sp2 = self.__p2
        _op1, _op2 = obj.getEndpoints()
        if ((_sp1 == _op1) and (_sp2 == _op2)):
            _val = False
        elif ((_sp1 == _op2) and (_sp2 == _op1)):
            _val = False
        else:
            _val = True
        return _val

    def getValues(self):
        _pids = entity.Entity.getValues(self)
        _gv = self.getGraphicValues(Segment.__defstyle)
        return 'segment', _pids, _gv, self.__p1.getID(), self.__p2.getID()

    def getEndpoints(self):
        """Get the endpoints of the Segment.

getEndpoints()

This function returns a tuple containing the two Point objects
that are the endpoints of the segment.
        """
        return self.__p1, self.__p2

    def getP1(self):
        """Return the first endpoint Point of the Segment.

getP1()
        """
        return self.__p1

    def setP1(self, p):
        """Set the first endpoint Point of the Segment.

setP1(p)
        """
        if self.isLocked():
            raise RuntimeError, "Setting endpoint not allowed - object locked."
        if not isinstance(p, point.Point):
            raise TypeError, "Invalid P1 endpoint: " + str(p)
        if p is self.__p2:
            raise ValueError, "Segments cannot have identical endpoints."
        _pt = self.__p1
        if _pt is not p:
            _pt.disconnect(self)
            _pt.freeUser(self)
            self.__p1 = p
            self.sendMessage('endpoint_changed', _pt, p)
            p.storeUser(self)
            p.connect('moved', self, Segment.movePoint)
            if abs(_pt.x - p.x) > 1e-10 or abs(_pt.y - p.y) > 1e-10:
                _x, _y = self.__p2.getCoords()
                self.sendMessage('moved', _pt.x, _pt.y, _x, _y)
            self.modified()

    p1 = property(getP1, setP1, None, "First endpoint of the Segment.")

    def getP2(self):
        """Return the second endpoint Point of the Segment.

getP2()
        """
        return self.__p2

    def setP2(self, p):
        """Set the second endpoint Point of the Segment.

setP2(p)
        """
        if self.isLocked():
            raise RuntimeError, "Setting endpoint not allowed - object locked."
        if not isinstance(p, point.Point):
            raise TypeError, "Invalid Point for p2 endpoint: " + str(p)
        if p is self.__p1:
            raise ValueError, "Segments cannot have identical endpoints."
        _pt = self.__p2
        if _pt is not p:
            _pt.disconnect(self)
            _pt.freeUser(self)
            self.__p2 = p
            self.sendMessage('endpoint_changed', _pt, p)
            p.storeUser(self)
            p.connect('moved', self, Segment.movePoint)
            if abs(_pt.x - p.x) > 1e-10 or abs(_pt.y - p.y) > 1e-10:
                _x, _y = self.__p1.getCoords()
                self.sendMessage('moved', _x, _y, _pt.x, _pt.y)
            self.modified()

    p2 = property(getP2, setP2, None, "Second endpoint of the Segment.")

    def move(self, dx, dy):
        """Move a Segment.

move(dx, dy)

The first argument gives the x-coordinate displacement,
and the second gives the y-coordinate displacement. Both
values should be floats.
        """
        if self.isLocked() or self.__p1.isLocked() or self.__p2.isLocked():
            raise RuntimeError, "Moving Segment not allowed - object locked."
        _dx = dx
        if not isinstance(_dx, float):
            _dx = float(dx)
        _dy = dy
        if not isinstance(_dy, float):
            _dy = float(dy)
        if abs(_dx) > 1e-10 or abs(_dy) > 1e-10:
            _x1, _y1 = self.__p1.getCoords()
            _x2, _y2 = self.__p2.getCoords()
            self.ignore('moved')
            try:
                self.__p1.move(_dx, _dy)
                self.__p2.move(_dx, _dy)
            finally:
                self.receive('moved')
            self.sendMessage('moved', _x1, _y1, _x2, _y2)
            self.modified()

    def length(self):
        """Return the length of the Segment.

length()
        """
        return self.__p1 - self.__p2

    def getCoefficients(self):
        """Express the line segment as a function ax + by + c = 0

getCoefficients()

This method returns a tuple of three floats: (a, b, c)
        """
        _x1, _y1 = self.__p1.getCoords()
        _x2, _y2 = self.__p2.getCoords()
        _a = _y2 - _y1
        _b = _x1 - _x2
        _c = (_x2 * _y1) - (_x1 * _y2)
        return _a, _b, _c

    def getProjection(self, x, y):
        """Find the projection point of some coordinates on the Segment.

getProjection(x, y)

Arguments "x" and "y" should be float values.
        """
        _x = x
        if not isinstance(_x, float):
            _x = float(x)
        _y = y
        if not isinstance(_y, float):
            _y = float(y)
        _x1, _y1 = self.__p1.getCoords()
        _x2, _y2 = self.__p2.getCoords()
        _sqlen = pow((_x2 - _x1), 2) +  pow((_y2 - _y1), 2)
        if _sqlen < 1e-10: # coincident points
            return None
        _rn = ((_x - _x1) * (_x2 - _x1)) + ((_y - _y1) * (_y2 - _y1))
        _r = _rn/_sqlen
        if _r < 0.0 or _r > 1.0:
            return None
        _px = _x1 + _r * (_x2 - _x1)
        _py = _y1 + _r * (_y2 - _y1)
        return _px, _py

    def mapCoords(self, x, y, tol=tolerance.TOL):
        """Return the nearest Point on the Segment to a coordinate pair.

mapCoords(x, y[, tol])

The function has two required arguments:

x: A Float value giving the "x" coordinate
y: A Float value giving the "y" coordinate

There is a single optional argument:

tol: A float value equal or greater than 0.

This function is used to map a possibly near-by coordinate pair to an
actual Point on the Segment. If the distance between the actual
Point and the coordinates used as an argument is less than the tolerance,
the actual Point is returned. Otherwise, this function returns None.
        """
        _x = x
        if not isinstance(_x, float):
            _x = float(x)
        _y = y
        if not isinstance(_y, float):
            _y = float(y)
        _t = tolerance.toltest(tol)
        _x1, _y1 = self.__p1.getCoords()
        _x2, _y2 = self.__p2.getCoords()
        return util.map_coords(_x, _y, _x1, _y1, _x2, _y2, _t)

    def inRegion(self, xmin, ymin, xmax, ymax, fully=False):
        """Return whether or not a Segment exists within a region.

inRegion(xmin, ymin, xmax, ymax[, fully])

The four arguments define the boundary of an area, and the
method returns True if the Segment lies within that area. If
the optional argument fully is used and is True, then both
endpoints of the Segment must lie within the boundary.
Otherwise, the method returns False.
        """
        _xmin = xmin
        if not isinstance(_xmin, float):
            _xmin = float(xmin)
        _ymin = ymin
        if not isinstance(_ymin, float):
            _ymin = float(ymin)
        _xmax = xmax
        if not isinstance(_xmax, float):
            _xmax = float(xmax)
        if _xmax < _xmin:
            raise ValueError, "Illegal values: xmax < xmin"
        _ymax = ymax
        if not isinstance(_ymax, float):
            _ymax = float(ymax)
        if _ymax < _ymin:
            raise ValueError, "Illegal values: ymax < ymin"
        if fully is not True and fully is not False:
            raise ValueError, "Invalid flag: " + str(fully)
        _x1, _y1 = self.__p1.getCoords()
        _x2, _y2 = self.__p2.getCoords()
        _pxmin = min(_x1, _x2)
        _pymin = min(_y1, _y2)
        _pxmax = max(_x1, _x2)
        _pymax = max(_y1, _y2)
        if ((_pxmax < _xmin) or
            (_pymax < _ymin) or
            (_pxmin > _xmax) or
            (_pymin > _ymax)):
            return False
        if fully:
            if ((_pxmin > _xmin) and
                (_pymin > _ymin) and
                (_pxmax < _xmax) and
                (_pymax < _ymax)):
                return True
            return False
        return util.in_region(_x1, _y1, _x2, _y2, _xmin, _ymin, _xmax, _ymax)

    def clipToRegion(self, xmin, ymin, xmax, ymax):
        """Clip the Segment using the Liang-Barsky Algorithm.

clipToRegion(xmin, ymin, xmax, ymax)
        """
        _xmin = xmin
        if not isinstance(_xmin, float):
            _xmin = float(xmin)
        _ymin = ymin
        if not isinstance(_ymin, float):
            _ymin = float(ymin)
        _xmax = xmax
        if not isinstance(_xmax, float):
            _xmax = float(xmax)
        if _xmax < _xmin:
            raise ValueError, "Illegal values: xmax < xmin"
        _ymax = ymax
        if not isinstance(_ymax, float):
            _ymax = float(ymax)
        if _ymax < _ymin:
            raise ValueError, "Illegal values: ymax < ymin"
        _x1, _y1 = self.__p1.getCoords()
        _x2, _y2 = self.__p2.getCoords()
        #
        # simple tests to reject line
        #
        if ((max(_x1, _x2) < _xmin) or
            (max(_y1, _y2) < _ymin) or
            (min(_x1, _x2) > _xmax) or
            (min(_y1, _y2) > _ymax)):
            return None
        #
        # simple tests to accept line
        #
        _coords = None
        if (_xmin < _x1 < _xmax and
            _xmin < _x2 < _xmax and
            _ymin < _y1 < _ymax and
            _ymin < _y2 < _ymax):
            _coords = (_x1, _y1, _x2, _y2)
        else:
            #
            # the Segment can be parameterized as
            #
            # x = u * (x2 - x1) + x1
            # y = u * (y2 - y1) + y1
            #
            # for u = 0, x => x1, y => y1
            # for u = 1, x => x2, y => y2
            #
            # The following is the Liang-Barsky Algorithm
            # for segment clipping
            #
            _dx = _x2 - _x1
            _dy = _y2 - _y1
            _P = [-_dx, _dx, -_dy, _dy]
            _q = [(_x1 - _xmin), (_xmax - _x1), (_y1 - _ymin), (_ymax - _y1)]
            _u1 = 0.0
            _u2 = 1.0
            _valid = True
            for _i in range(4):
                _pi = _P[_i]
                _qi = _q[_i]
                if abs(_pi) < 1e-10:
                    if _qi < 0.0:
                        _valid = False
                        break
                else:
                    _r = _qi/_pi
                    if _pi < 0.0:
                        if _r > _u2:
                            _valid = False
                            break
                        if _r > _u1:
                            _u1 = _r
                    else:
                        if _r < _u1:
                            _valid = False
                            break
                        if _r < _u2:
                            _u2 = _r
            if _valid:
                _coords = (((_u1 * _dx) + _x1),
                           ((_u1 * _dy) + _y1),
                           ((_u2 * _dx) + _x1),
                           ((_u2 * _dy) + _y1))
        return _coords

    def movePoint(self, p, *args):
        _alen = len(args)
        if _alen < 2:
            raise ValueError, "Invalid argument count: %d" % _alen
        _x = args[0]
        if not isinstance(_x, float):
            _x = float(args[0])
        _y = args[1]
        if not isinstance(_y, float):
            _y = float(args[1])
        _p1 = self.__p1
        _p2 = self.__p2
        if p is _p1:
            _x1 = _x
            _y1 = _y
            _x2, _y2 = _p2.getCoords()
        elif p is _p2:
            _x1, _y1 = _p1.getCoords()
            _x2 = _x
            _y2 = _y
        else:
            raise ValueError, "Unexpected Segment endpoint: " + `p`
        self.sendMessage('moved', _x1, _y1, _x2, _y2)
        self.modified()

    def clone(self):
        """Create an identical copy of a Segment.

clone()
        """
        _cp1 = self.__p1.clone()
        _cp2 = self.__p2.clone()
        _st = self.getStyle()
        _lt = self.getLinetype()
        _col = self.getColor()
        _th = self.getThickness()
        return Segment(_cp1, _cp2, _st, _lt, _col, _th)

    def sendsMessage(self, m):
        if m in Segment.messages:
            return True
        return graphicobject.GraphicObject.sendsMessage(self, m)

#
# Quadtree Segment storage
#

class SegmentQuadtree(quadtree.Quadtree):
    def __init__(self):
        quadtree.Quadtree.__init__(self)

    def getNodes(self, *args):
        _alen = len(args)
        if _alen != 4:
            raise ValueError, "Expected 4 arguments, got %d" % _alen
        _x1, _y1, _x2, _y2 = args
        if not isinstance(_x1, float):
            _x1 = float(args[0])
        if not isinstance(_y1, float):
            _y1 = float(args[1])
        if not isinstance(_x2, float):
            _x2 = float(args[2])
        if not isinstance(_y2, float):
            _y2 = float(args[3])
        _sxmin = min(_x1, _x2)
        _sxmax = max(_x1, _x2)
        _symin = min(_y1, _y2)
        _symax = max(_y1, _y2)
        _nodes = [self.getTreeRoot()]
        while len(_nodes):
            _node = _nodes.pop()
            _xmin, _ymin, _xmax, _ymax = _node.getBoundary()
            if ((_sxmin > _xmax) or
                (_sxmax < _xmin) or
                (_symin > _ymax) or
                (_symax < _ymin)):
                continue
            if _node.hasSubnodes():
                _xmid = (_xmin + _xmax)/2.0
                _ymid = (_ymin + _ymax)/2.0
                _ne = _nw = _sw = _se = True
                if _sxmax < _xmid: # seg on left side
                    _ne = _se = False
                if _sxmin > _xmid: # seg on right side
                    _nw = _sw = False
                if _symax < _ymid: # seg below
                    _nw = _ne = False
                if _symin > _ymid: # seg above
                    _sw = _se = False
                if _ne:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NENODE))
                if _nw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NWNODE))
                if _sw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SWNODE))
                if _se:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SENODE))
            else:
                yield _node

    def addObject(self, obj):
        if not isinstance(obj, Segment):
            raise TypeError, "Invalid Segment object: " + `obj`
        if obj in self:
            return
        _p1, _p2 = obj.getEndpoints()
        _x1, _y1 = _p1.getCoords()
        _x2, _y2 = _p2.getCoords()
        _bounds = self.getTreeRoot().getBoundary()
        _xmin = _ymin = _xmax = _ymax = None
        _sxmin = min(_x1, _x2)
        _sxmax = max(_x1, _x2)
        _symin = min(_y1, _y2)
        _symax = max(_y1, _y2)
        _resize = False
        if _bounds is None: # first node in tree
            _resize = True
            _xmin = _sxmin - 1.0
            _ymin = _symin - 1.0
            _xmax = _sxmax + 1.0
            _ymax = _symax + 1.0
        else:
            _xmin, _ymin, _xmax, _ymax = _bounds
            if _sxmin < _xmin:
                _xmin = _sxmin - 1.0
                _resize = True
            if _sxmax > _xmax:
                _xmax = _sxmax + 1.0
                _resize = True
            if _symin < _ymin:
                _ymin = _symin - 1.0
                _resize = True
            if _symax > _ymax:
                _ymax = _symax + 1.0
                _resize = True
        if _resize:
            self.resize(_xmin, _ymin, _xmax, _ymax)
        for _node in self.getNodes(_x1, _y1, _x2, _y2):
            _xmin, _ymin, _xmax, _ymax = _node.getBoundary()
            if obj.inRegion(_xmin, _ymin, _xmax, _ymax):
                _node.addObject(obj)
        quadtree.Quadtree.addObject(self, obj)
        obj.connect('moved', self, SegmentQuadtree.moveSegment)

    def old_addObject(self, obj):
        if not isinstance(obj, Segment):
            raise TypeError, "Invalid Segment object: " + `obj`
        if obj in self:
            return
        _p1, _p2 = obj.getEndpoints()
        _x1, _y1 = _p1.getCoords()
        _x2, _y2 = _p2.getCoords()
        _node = self.getTreeRoot()
        _bounds = _node.getBoundary()
        _xmin = _ymin = _xmax = _ymax = None
        _sxmin = min(_x1, _x2)
        _sxmax = max(_x1, _x2)
        _symin = min(_y1, _y2)
        _symax = max(_y1, _y2)
        _resize = False
        if _bounds is None: # first node in tree
            _resize = True
            _xmin = _sxmin - 1.0
            _ymin = _symin - 1.0
            _xmax = _sxmax + 1.0
            _ymax = _symax + 1.0
        else:
            _xmin, _ymin, _xmax, _ymax = _bounds
            if _sxmin < _xmin:
                _xmin = _sxmin - 1.0
                _resize = True
            if _sxmax > _xmax:
                _xmax = _sxmax + 1.0
                _resize = True
            if _symin < _ymin:
                _ymin = _symin - 1.0
                _resize = True
            if _symax > _ymax:
                _ymax = _symax + 1.0
                _resize = True
        if _resize:
            self.resize(_xmin, _ymin, _xmax, _ymax)
            _node = self.getTreeRoot()
        _nodes = [_node]
        while len(_nodes):
            _node = _nodes.pop()
            _xmin, _ymin, _xmax, _ymax = _node.getBoundary()
            if ((_sxmin > _xmax) or
                (_sxmax < _xmin) or
                (_symin > _ymax) or
                (_symax < _ymin)):
                continue
            if _node.hasSubnodes():
                _xmid = (_xmin + _xmax)/2.0
                _ymid = (_ymin + _ymax)/2.0
                _ne = _nw = _sw = _se = True
                if _sxmax < _xmid: # seg on left side
                    _ne = _se = False
                if _sxmin > _xmid: # seg on right side
                    _nw = _sw = False
                if _symax < _ymid: # seg below
                    _nw = _ne = False
                if _symin > _ymid: # seg above
                    _sw = _se = False
                if _ne:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NENODE))
                if _nw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NWNODE))
                if _sw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SWNODE))
                if _se:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SENODE))
            else:
                if obj.inRegion(_xmin, _ymin, _xmax, _ymax):
                    _node.addObject(obj)
        quadtree.Quadtree.addObject(self, obj)
        obj.connect('moved', self, SegmentQuadtree.moveSegment)

    def delObject(self, obj):
        if obj not in self:
            return
        _p1, _p2 = obj.getEndpoints()
        _x1, _y1 = _p1.getCoords()
        _x2, _y2 = _p2.getCoords()
        _pdict = {}
        for _node in self.getNodes(_x1, _y1, _x2, _y2):
            _node.delObject(obj)
            _parent = _node.getParent()
            if _parent is not None:
                _pid = id(_parent)
                if _pid not in _pdict:
                    _pdict[_pid] = _parent
        quadtree.Quadtree.delObject(self, obj)
        obj.disconnect(self)
        for _parent in _pdict.values():
            self.purgeSubnodes(_parent)

    def old_delObject(self, obj):
        if obj not in self:
            return
        _p1, _p2 = obj.getEndpoints()
        _x1, _y1 = _p1.getCoords()
        _x2, _y2 = _p2.getCoords()
        _xmin = _ymin = _xmax = _ymax = None
        _sxmin = min(_x1, _x2)
        _sxmax = max(_x1, _x2)
        _symin = min(_y1, _y2)
        _symax = max(_y1, _y2)
        _pdict = {}
        _nodes = [self.getTreeRoot()]
        while len(_nodes):
            _node = _nodes.pop()
            _xmin, _ymin, _xmax, _ymax = _node.getBoundary()
            if ((_sxmin > _xmax) or
                (_sxmax < _xmin) or
                (_symin > _ymax) or
                (_symax < _ymin)):
                continue
            if _node.hasSubnodes():
                _xmid = (_xmin + _xmax)/2.0
                _ymid = (_ymin + _ymax)/2.0
                _ne = _nw = _sw = _se = True
                if _sxmax < _xmid: # seg on left side
                    _ne = _se = False
                if _sxmin > _xmid: # seg on right side
                    _nw = _sw = False
                if _symax < _ymid: # seg below
                    _nw = _ne = False
                if _symin > _ymid: # seg above
                    _sw = _se = False
                if _ne:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NENODE))
                if _nw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NWNODE))
                if _sw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SWNODE))
                if _se:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SENODE))
            else:
                _node.delObject(obj) # seg may not be in the node ...
                _parent = _node.getParent()
                if _parent is not None:
                    _pid = id(_parent)
                    if _pid not in _pdict:
                        _pdict[_pid] = _parent
        quadtree.Quadtree.delObject(self, obj)
        obj.disconnect(self)
        for _parent in _pdict.values():
            self.purgeSubnodes(_parent)

    def new_find(self, *args):
        _alen = len(args)
        if _alen < 4:
            raise ValueError, "Invalid argument count: %d" % _alen
        _x1 = args[0]
        if not isinstance(_x1, float):
            _x1 = float(args[0])
        _y1 = args[1]
        if not isinstance(_y1, float):
            _y1 = float(args[1])
        _x2 = args[2]
        if not isinstance(_x2, float):
            _x2 = float(args[2])
        _y2 = args[3]
        if not isinstance(_y2, float):
            _y2 = float(args[3])
        _t = tolerance.TOL
        if _alen > 4:
            _t = tolerance.toltest(args[4])
        if not len(self):
            return None
        _sdict = {}
        _segs = []
        for _node in self.getNodes(_x1, _y1, _x2, _y2):
            _xmin, _ymin, _xmax, _ymax = _node.getBoundary()
            for _s in _node.getObjects():
                _sid = id(_s)
                if _sid not in _sdict:
                    _p1, _p2 = _s.getEndpoints()
                    _px1, _py1 = _p1.getCoords()
                    _px2, _py2 = _p2.getCoords()
                    if ((abs(_px1 - _x1) < _t) and
                        (abs(_py1 - _y1) < _t) and
                        (abs(_px2 - _x2) < _t) and
                        (abs(_py2 - _y2) < _t)):
                        _segs.append(_s)
                    elif ((abs(_px2 - _x1) < _t) and
                          (abs(_py2 - _y1) < _t) and
                          (abs(_px1 - _x2) < _t) and
                          (abs(_py1 - _y2) < _t)):
                        _segs.append(_s)
                    _sdict[_sid] = True
        return _segs

    def find(self, *args):
        _alen = len(args)
        if _alen < 4:
            raise ValueError, "Invalid argument count: %d" % _alen
        _x1 = args[0]
        if not isinstance(_x1, float):
            _x1 = float(args[0])
        _y1 = args[1]
        if not isinstance(_y1, float):
            _y1 = float(args[1])
        _x2 = args[2]
        if not isinstance(_x2, float):
            _x2 = float(args[2])
        _y2 = args[3]
        if not isinstance(_y2, float):
            _y2 = float(args[3])
        _t = tolerance.TOL
        if _alen > 4:
            _t = tolerance.toltest(args[4])
        if not len(self):
            return None
        _sxmin = min(_x1, _x2)
        _sxmax = max(_x1, _x2)
        _symin = min(_y1, _y2)
        _symax = max(_y1, _y2)
        _nodes = [self.getTreeRoot()]
        _sdict = {}
        _bailout = False
        _seg = None
        while len(_nodes):
            _node = _nodes.pop()
            _xmin, _ymin, _xmax, _ymax = _node.getBoundary()
            if ((_sxmin > _xmax) or
                (_sxmax < _xmin) or
                (_symin > _ymax) or
                (_symax < _ymin)):
                continue
            if _node.hasSubnodes():
                _xmid = (_xmin + _xmax)/2.0
                _ymid = (_ymin + _ymax)/2.0
                _ne = _nw = _sw = _se = True
                if _sxmax < (_xmid - _t): # seg on left side
                    _ne = _se = False
                if _sxmin > (_xmid + _t): # seg on right side
                    _nw = _sw = False
                if _symax < (_ymid - _t): # seg below
                    _nw = _ne = False
                if _symin > (_ymid + _t): # seg above
                    _sw = _se = False
                if _ne:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NENODE))
                if _nw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NWNODE))
                if _sw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SWNODE))
                if _se:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SENODE))
            else:
                for _s in _node.getObjects():
                    _sid = id(_s)
                    if _sid not in _sdict:
                        _p1, _p2 = _s.getEndpoints()
                        _px, _py = _p1.getCoords()
                        if ((abs(_px - _x1) < _t) and (abs(_py - _y1) < _t)):
                            _px, _py = _p2.getCoords()
                            if ((abs(_px - _x2) < _t) and
                                (abs(_py - _y2) < _t)):
                                _bailout = True
                                _seg = _s
                                break
                        _px, _py = _p2.getCoords()
                        if ((abs(_px - _x1) < _t) and (abs(_py - _y1) < _t)):
                            _px, _py = _p1.getCoords()
                            if ((abs(_px - _x2) < _t) and
                                (abs(_py - _y2) < _t)):
                                _bailout = True
                                _seg = _s
                                break
                        _sdict[_sid] = True
            if _bailout:
                break
        return _seg

    def moveSegment(self, obj, *args):
        if obj not in self:
            raise ValueError, "Segment not stored in Quadtree: " + `obj`
        _alen = len(args)
        if _alen < 4:
            raise ValueError, "Invalid argument count: %d" % _alen
        _x1 = args[0]
        if not isinstance(_x1, float):
            _x1 = float(args[0])
        _y1 = args[1]
        if not isinstance(_y1, float):
            _y1 = float(args[1])
        _x2 = args[2]
        if not isinstance(_x2, float):
            _x2 = float(args[2])
        _y2 = args[3]
        if not isinstance(_y2, float):
            _y2 = float(args[3])
        for _node in self.getNodes(_x1, _y1, _x2, _y2):
            _node.delObject(obj) # segment may not be in node ...
        quadtree.Quadtree.delObject(self, obj)
        obj.disconnect(self)
        self.addObject(obj)

    def old_moveSegment(self, obj, *args):
        if obj not in self:
            raise ValueError, "Segment not stored in Quadtree: " + `obj`
        _alen = len(args)
        if _alen < 4:
            raise ValueError, "Invalid argument count: %d" % _alen
        _x1 = args[0]
        if not isinstance(_x1, float):
            _x1 = float(args[0])
        _y1 = args[1]
        if not isinstance(_y1, float):
            _y1 = float(args[1])
        _x2 = args[2]
        if not isinstance(_x2, float):
            _x2 = float(args[2])
        _y2 = args[3]
        if not isinstance(_y2, float):
            _y2 = float(args[3])
        _sxmin = min(_x1, _x2)
        _sxmax = max(_x1, _x2)
        _symin = min(_y1, _y2)
        _symax = max(_y1, _y2)
        _nodes = [self.getTreeRoot()]
        while len(_nodes):
            _node = _nodes.pop()
            _xmin, _ymin, _xmax, _ymax = _node.getBoundary()
            if ((_sxmin > _xmax) or
                (_sxmax < _xmin) or
                (_symin > _ymax) or
                (_symax < _ymin)):
                continue
            if _node.hasSubnodes():
                _xmid = (_xmin + _xmax)/2.0
                _ymid = (_ymin + _ymax)/2.0
                _ne = _nw = _sw = _se = True
                if _sxmax < _xmid: # seg on left side
                    _ne = _se = False
                if _sxmin > _xmid: # seg on right side
                    _nw = _sw = False
                if _symax < _ymid: # seg below
                    _nw = _ne = False
                if _symin > _ymid: # seg above
                    _sw = _se = False
                if _ne:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NENODE))
                if _nw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NWNODE))
                if _sw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SWNODE))
                if _se:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SENODE))
            else:
                _node.delObject(obj) # segment may not be in node ...
        quadtree.Quadtree.delObject(self, obj)
        obj.disconnect(self)
        self.addObject(obj)

    def getClosest(self, x, y, tol=tolerance.TOL):
        _x = x
        if not isinstance(_x, float):
            _x = float(x)
        _y = y
        if not isinstance(_y, float):
            _y = float(y)
        _t = tolerance.toltest(tol)
        _seg = _tsep = None
        _bailout = False
        _sdict = {}
        _nodes = [self.getTreeRoot()]
        while len(_nodes):
            _node = _nodes.pop()
            _xmin, _ymin, _xmax, _ymax = _node.getBoundary()
            if ((_x < (_xmin - _t)) or
                (_x > (_xmax + _t)) or
                (_y < (_ymin - _t)) or
                (_y > (_ymax + _t))):
                continue
            if _node.hasSubnodes():
                _nodes.extend(_node.getSubnodes())
            else:
                for _s in _node.getObjects():
                    _sid = id(_s)
                    if _sid not in _sdict:
                        _p1, _p2 = _s.getEndpoints()
                        _px, _py = _p1.getCoords()
                        if ((abs(_px - _x) < 1e-10) and
                            (abs(_py - _y) < 1e-10)):
                            _seg = _s
                            _bailout = True
                            break
                        _px, _py = _p2.getCoords()
                        if ((abs(_px - _x) < 1e-10) and
                            (abs(_py - _y) < 1e-10)):
                            _seg = _s
                            _bailout = True
                            break
                        _sdict[_sid] = True
                    _pt = _s.mapCoords(_x, _y, _t)
                    if _pt is not None:
                        _px, _py = _pt
                        _sep = math.hypot((_px - _x), (_py - _y))
                        if _tsep is None:
                            _tsep = _sep
                            _seg = _s
                        else:
                            if _sep < _tsep:
                                _tsep = _sep
                                _seg = _s
            if _bailout:
                break
        return _seg

    def getInRegion(self, xmin, ymin, xmax, ymax):
        _xmin = xmin
        if not isinstance(_xmin, float):
            _xmin = float(xmin)
        _ymin = ymin
        if not isinstance(_ymin, float):
            _ymin = float(ymin)
        _xmax = xmax
        if not isinstance(_xmax, float):
            _xmax = float(xmax)
        if _xmax < _xmin:
            raise ValueError, "Illegal values: xmax < xmin"
        _ymax = ymax
        if not isinstance(_ymax, float):
            _ymax = float(ymax)
        if _ymax < _ymin:
            raise ValueError, "Illegal values: ymax < ymin"
        _segs = []
        if not len(self):
            return _segs
        _nodes = []
        _nodes.append(self.getTreeRoot())
        _sdict = {}
        while len(_nodes):
            _node = _nodes.pop()
            if _node.hasSubnodes():
                for _subnode in _node.getSubnodes():
                    _sxmin, _symin, _sxmax, _symax = _subnode.getBoundary()
                    if ((_sxmin > _xmax) or
                        (_symin > _ymax) or
                        (_sxmax < _xmin) or
                        (_symax < _ymin)):
                        continue
                    _nodes.append(_subnode)
            else:
                for _seg in _node.getObjects():
                    _sid = id(_seg)
                    if _sid not in _sdict:
                        if _seg.inRegion(_xmin, _ymin, _xmax, _ymax):
                            _segs.append(_seg)
                        _sdict[_sid] = True
        return _segs

#
# Segment history class
#

class SegmentLog(graphicobject.GraphicObjectLog):
    def __init__(self, s):
        if not isinstance(s, Segment):
            raise TypeError, "Invalid segment: " + `s`
        graphicobject.GraphicObjectLog.__init__(self, s)
        s.connect('endpoint_changed', self, SegmentLog.endpointChange)

    def endpointChange(self, s, *args):
        _alen = len(args)
        if _alen < 2:
            raise ValueError, "Invalid argument count: %d" % _alen
        _old = args[0]
        if not isinstance(_old, point.Point):
            raise TypeError, "Invalid old endpoint: " + `_old`
        _oid = _old.getID()
        _new = args[1]
        if not isinstance(_new, point.Point):
            raise TypeError, "Invalid new endpoint: " + `_new`
        _nid = _new.getID()
        self.saveUndoData('endpoint_changed', _oid, _nid)

    def execute(self, undo, *args):
        if undo is not True and undo is not False:
            raise ValueError, "Invalid undo value: " + str(undo)
        _alen = len(args)
        if len(args) == 0:
            raise ValueError, "No arguments to execute()"
        _s = self.getObject()
        _p1, _p2 = _s.getEndpoints()
        _op = args[0]
        if _op == 'endpoint_changed':
            if _alen < 3:
                raise ValueError, "Invalid argument count: %d" % _alen
            _oid = args[1]
            _nid = args[2]
            if undo:
                self.saveRedoData('endpoint_changed', _oid, _nid)
                self.ignore('endpoint_changed')
                try:
                    _parent = _s.getParent()
                    if _parent is None:
                        raise ValueError, "Segment has no parent - cannot undo"
                    _op = _parent.getObject(_oid)
                    if _op is None or not isinstance(_op, point.Point):
                        raise ValueError, "Old endpoint missing: id=%d" % _oid
                    _x1, _y1 = _p1.getCoords()
                    _x2, _y2 = _p2.getCoords()
                    _moved = False
                    _s.startUndo(True)
                    try:
                        if _p1.getID() == _nid:
                            _s.setP1(_op)
                            if (abs(_op.x - _x1) > 1e-10 or
                                abs(_op.y - _y1) > 1e-10):
                                _moved = True
                        elif _p2.getID() == _nid:
                            _s.setP2(_op)
                            if (abs(_op.x - _x2) > 1e-10 or
                                abs(_op.y - _y2) > 1e-10):
                                _moved = True
                        else:
                            raise ValueError, "Unexpected endpoint ID: %d" % _nid
                    finally:
                        _s.endUndo()
                    if _moved:
                        _s.sendMessage('moved', _x1, _y1, _x2, _y2)
                finally:
                    self.receive('endpoint_changed')
            else:
                self.saveUndoData('endpoint_changed', _oid, _nid)
                _parent = _s.getParent()
                if _parent is None:
                    raise ValueError, "Segment has no parent - cannot undo"
                _np = _parent.getObject(_nid)
                if _np is None or not isinstance(_np, point.Point):
                    raise ValueError, "New endpoint missing: id=%d" % _nid
                self.ignore('endpoint_changed')
                try:
                    if _p1.getID() == _oid:
                        _s.setP1(_np)
                    elif _p2.getID() == _oid:
                        _s.setP2(_np)
                    else:
                        raise ValueError, "Unexpected endpoint ID: %d" % _oid
                finally:
                    self.receive('endpoint_changed')
        else:
            raise ValueError, "Unexpected operation: %s" % _op



#
# Copyright (c) 2004 Art Haas
#
# This file is part of PythonCAD.
#
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# _The_ base class for all objects used in PythonCAD
#

import message
import logger

class Entity(message.Messenger):
    """The base class for PythonCAD entities.

An entity has the following properties:

parent: The parent of an entity - another entity or None

An entity has the following methods:

isLocked(): Returns the locked state of an entity.
lock()/unlock(): Set/Unset a flag indicating the entity should not be altered.
isModified(): Returns the flag indicating the entity has been modified.
modified()/reset(): Set/Unset a flag indicating the entity modified state.
isVisible(): Returns the visibity of the entity.
hide(): Make the entity invisible.
show()/expose(): Make the entity visible.
{set/get}Parent(): Set/Get the parent entity of an entity.
{add/del}Child(): Add/Remove a child entity to/from another entity.
hasChild(): Test if one entity is a child of another.
hasChildren(): Test if an entity has any children entities.
getChildren(): Return a list of children entities for an entity.
    """
    messages = {
        'locked' : True,
        'unlocked' : True,
        'modified' : True,
        'reset' : True,
        'reparented' : True,
        'hidden' : True,
        'exposed' : True,
        'added_child' : True,
        'removed_child' : True,
        }

    idcount = 0

    def __init__(self, **kw):
        _parent = _id = None
        if 'parent' in kw:
            _parent = kw['parent']
            if _parent is not None and not isinstance(_parent, Entity):
                raise TypeError, "Invalid parent: " + `_parent`
        if 'id' in kw:
            _id = kw['id']
            if not isinstance(_id, int):
                raise TypeError, "Invalid ID: " + `_id`
            if _id < 0 or _id >= Entity.idcount:
                raise ValueError, "Invalid ID: %d" % _id
        if _id is None:
            _id = Entity.idcount
            Entity.idcount = Entity.idcount + 1
        message.Messenger.__init__(self)
        self.__locked = False
        self.__modified = False
        self.__visible = True
        self.__parent = None
        self.__children = None
        self.__log = None
        self.__undoing = False
        self.__id = _id
        if _parent is not None:
            Entity.setParent(self, _parent)

    def getValues(self):
        _pid = None
        if self.__parent is not None:
            _pid = self.__parent.getID()
        return self.__id, _pid

    values = property(getValues, None, None, "Entity values")

    def getID(self):
        return self.__id

    id = property(getID, None, None, "Entity ID")

    def isLocked(self):
        return self.__locked

    def lock(self):
        if not self.__locked:
            self.__locked = True
        self.sendMessage('locked')

    def unlock(self):
        if self.__locked:
            self.__locked = False
        self.sendMessage('unlocked')

    def isModified(self):
        return self.__modified

    def modified(self):
        if not self.__modified:
            self.__modified = True
        # if not self.__undoing:
            # self.sendMessage('modified') # always send this message
        self.sendMessage('modified')

    def reset(self):
        if self.__modified:
            self.__modified = False
        if not self.__undoing:
            self.sendMessage('reset') # always send this message

    def isVisible(self):
        return self.__visible

    def hide(self):
        if self.__visible:
            self.__visible = False
        self.sendMessage('hidden')

    def show(self):
        if not self.__visible:
            self.__visible = True
        self.sendMessage('exposed')

    def expose(self):
        Entity.show(self)

    def canParent(self, obj):
        """Test if an Entity can be the parent of another Entity.

canParent(obj)

Subclasses should override this method if the ability to be
the parent of another entity needs refinement.
        """
        return True

    def setParent(self, parent):
        if parent is not None:
            if not isinstance(parent, Entity):
                raise TypeError, "Parent is not an Entity: " + `parent`
            if not parent.canParent(self):
                raise ValueError, "Invalid parent for Entity: " + `parent`
        _oldparent = self.__parent
        if _oldparent is not parent:
            if _oldparent is not None:
                Entity.delChild(_oldparent, self)
            self.__parent = parent
            if parent is not None:
                Entity.addChild(parent, self)
            self.sendMessage('reparented', _oldparent)
            self.modified()

    def getParent(self):
        return self.__parent

    parent = property(getParent, setParent, None, "Parent of an Entity.")

    def addChild(self, child):
        if not isinstance(child, Entity):
            raise TypeError, "Invalid child: " + `child`
        if child.getParent() is not self:
            raise ValueError, "Invalid parent: " + `child`
        if self.__children is None:
            self.__children = []
        for _child in self.__children:
            if _child is child:
                raise ValueError, "Child entity already stored: " + `child`
        self.__children.append(child)
        self.sendMessage('added_child', child)
        self.modified()

    def delChild(self, child):
        if child.getParent() is not self:
            raise ValueError, "Invalid parent for child: " + `child`
        if self.__children is None:
            raise ValueError, "Entity has no children: " + `self`
        _seen = False
        for _child in self.__children:
            if _child is _child:
                _seen = True
                break
        if not _seen:
            raise ValueError, "Child entity not stored: " + `child`
        self.__children.remove(child)
        if len(self.__children) == 0:
            self.__children = None
        self.sendMessage('removed_child', child)
        self.modified()

    def hasChild(self, child):
        return self.__children is not None and child in self.__children

    def hasChildren(self):
        return self.__children is not None

    def getChildren(self):
        if self.__children is not None:
            return self.__children[:]
        return []

    def sendsMessage(self, m):
        if m in Entity.messages:
            return True
        return message.Messenger.sendsMessage(self, m)

    def clone(self):
        return Entity(parent=self.__parent)

    def inRegion(self, xmin, ymin, xmax, ymax, all=False):
        return False

    def getLog(self):
        return self.__log

    def setLog(self, log):
        if log is not None:
            if not isinstance(log, logger.Logger):
                raise TypeError, "Invalid log object: " + `log`
            if self.__log is not None:
                raise ValueError, "Entity already contains a log."
        self.__log = log

    def delLog(self):
        if self.__log is not None:
            self.__log.clear()
        self.__log = None

    def undo(self):
        if self.__log is not None:
            self.__log.undo()

    def startUndo(self, mute=True):
        if mute:
            self.mute()
        self.__undoing = True

    def endUndo(self):
        if self.isMuted():
            self.unmute()
        self.__undoing = False

    def inUndo(self):
        return self.__undoing

    def redo(self):
        if self.__log is not None:
            self.__log.redo()
